@@ -1,6 +1,6 @@
 /*
-   Copyright (C) 2006-2012 Free Software Foundation, Inc.
-   Written by Roger While
+   Copyright (C) 2006-2012, 2013, 2017-2018 Free Software Foundation, Inc.
+   Written by Roger While, Ron Norman
 
    This file is part of GnuCOBOL.
 
@@ -15,11 +15,10 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GnuCOBOL.  If not, see <http://www.gnu.org/licenses/>.
+   along with GnuCOBOL.  If not, see <https://www.gnu.org/licenses/>.
 */
 
-#include "config.h"
-#include "defaults.h"
+#include <config.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -31,6 +30,12 @@
 #include "cobc.h"
 #include "tree.h"
 
+#ifdef COB_NO_UNALIGNED_ATTRIBUTE
+#define UNALIGNED_ATTRIBUTE ""
+#else
+#define UNALIGNED_ATTRIBUTE "__unaligned "
+#endif
+
 static void
 output_storage (const char *fmt, ...)
 {
@@ -47,6 +52,7 @@ output_storage (const char *fmt, ...)
 void
 cob_gen_optim (const enum cb_optim val)
 {
+	output_storage ("");
 	switch (val) {
 
 	case COB_SET_SCREEN:
@@ -75,6 +81,39 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("}");
 		return;
 
+	case COB_SET_REPORT:
+		output_storage ("static void COB_NOINLINE");
+		output_storage ("cob_set_report (cob_report *r, cob_file *pfile)");
+		output_storage ("{");
+		output_storage ("	r->report_file = pfile;");
+		output_storage ("}");
+		return;
+
+	case COB_SET_ML_TREE:
+		output_storage ("static void COB_NOINLINE");
+		output_storage ("cob_set_ml_attr (cob_ml_attr *attr, cob_field *name,");
+		output_storage ("		cob_field *value, unsigned int is_suppressed,");
+		output_storage ("		cob_ml_attr *sibling)");
+		output_storage ("{");
+		output_storage ("	attr->name = name;");
+		output_storage ("	attr->value = value;");
+		output_storage ("	attr->is_suppressed = is_suppressed;");
+		output_storage ("	attr->sibling = sibling;");
+		output_storage ("}");
+		output_storage ("static void COB_NOINLINE");
+		output_storage ("cob_set_ml_tree (cob_ml_tree *tree, cob_field *name, cob_ml_attr *attrs,");
+		output_storage ("		cob_field *content, unsigned int is_suppressed,");
+		output_storage ("		cob_ml_tree *children, cob_ml_tree *sibling)");
+		output_storage ("{");
+		output_storage ("       tree->name = name;");
+		output_storage ("	tree->attrs = attrs;");
+		output_storage ("	tree->content = content;");
+		output_storage ("	tree->is_suppressed = is_suppressed;");
+		output_storage ("	tree->children = children;");
+		output_storage ("	tree->sibling = sibling;");
+		output_storage ("}");
+		return;		
+
 	case COB_POINTER_MANIP:
 		output_storage ("static void COB_NOINLINE");
 		output_storage ("cob_pointer_manip (cob_field *f1, cob_field *f2, const unsigned int addsub)");
@@ -92,17 +131,18 @@ cob_gen_optim (const enum cb_optim val)
 
 	case COB_GET_NUMDISP:
 		output_storage ("static int COB_NOINLINE");
-		output_storage ("cob_get_numdisp (const void *data, const size_t size)");
+		output_storage ("cob_get_numdisp (const void *data, const int size)");
 		output_storage ("{");
 		output_storage ("	const unsigned char	*p;");
-		output_storage ("	size_t			n;");
-		output_storage ("	int    			 retval;");
+		output_storage ("	int			n;");
+		output_storage ("	int 			retval;");
 
 		output_storage ("	p = (const unsigned char *)data;");
 		output_storage ("	retval = 0;");
 		output_storage ("	for (n = 0; n < size; ++n, ++p) {");
 		output_storage ("		retval *= 10;");
-		output_storage ("		retval += (*p & 0x0F);");
+		output_storage ("		if (*p > '0' && *p <= '9')");
+		output_storage ("		    retval += (*p - '0');");
 		output_storage ("	}");
 		output_storage ("	return retval;");
 		output_storage ("}");
@@ -215,7 +255,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	if (unlikely(n < 0)) {");
 		output_storage ("		return 1;");
 		output_storage ("	}");
-		output_storage ("	val = *(unsigned short __unaligned *)p;");
+		output_storage ("	val = *(unsigned short " UNALIGNED_ATTRIBUTE "*)p;");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -226,7 +266,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("{");
 		output_storage ("	short	val;");
 
-		output_storage ("	val = *(short __unaligned *)p;");
+		output_storage ("	val = *(short " UNALIGNED_ATTRIBUTE "*)p;");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -240,7 +280,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	if (unlikely(n < 0)) {");
 		output_storage ("		return 1;");
 		output_storage ("	}");
-		output_storage ("	val = *(unsigned int __unaligned *)p;");
+		output_storage ("	val = *(unsigned int " UNALIGNED_ATTRIBUTE "*)p;");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -251,7 +291,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("{");
 		output_storage ("	int	val;");
 
-		output_storage ("	val = *(int __unaligned *)p;");
+		output_storage ("	val = *(int " UNALIGNED_ATTRIBUTE "*)p;");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -262,7 +302,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("{");
 		output_storage ("	cob_u64_t	val;");
 
-		output_storage ("	val = *(cob_u64_t __unaligned *)p;");
+		output_storage ("	val = *(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p;");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -273,7 +313,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("{");
 		output_storage ("	cob_s64_t	val;");
 
-		output_storage ("	val = *(cob_s64_t __unaligned *)p;");
+		output_storage ("	val = *(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p;");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -284,7 +324,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_add_align_u16 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(unsigned short __unaligned *)p += val;");
+		output_storage ("	*(unsigned short " UNALIGNED_ATTRIBUTE "*)p += val;");
 		output_storage ("}");
 		return;
 
@@ -292,7 +332,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_add_align_s16 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(short __unaligned *)p += val;");
+		output_storage ("	*(short " UNALIGNED_ATTRIBUTE "*)p += val;");
 		output_storage ("}");
 		return;
 
@@ -300,7 +340,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_add_align_u32 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(unsigned int __unaligned *)p += val;");
+		output_storage ("	*(unsigned int " UNALIGNED_ATTRIBUTE "*)p += val;");
 		output_storage ("}");
 		return;
 
@@ -308,7 +348,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_add_align_s32 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(int __unaligned *)p += val;");
+		output_storage ("	*(int " UNALIGNED_ATTRIBUTE "*)p += val;");
 		output_storage ("}");
 		return;
 
@@ -316,7 +356,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_add_align_u64 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(cob_u64_t __unaligned *)p += val;");
+		output_storage ("	*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p += val;");
 		output_storage ("}");
 		return;
 
@@ -324,7 +364,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_add_align_s64 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(cob_s64_t __unaligned *)p += val;");
+		output_storage ("	*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p += val;");
 		output_storage ("}");
 		return;
 
@@ -334,7 +374,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_sub_align_u16 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(unsigned short __unaligned *)p -= val;");
+		output_storage ("	*(unsigned short " UNALIGNED_ATTRIBUTE "*)p -= val;");
 		output_storage ("}");
 		return;
 
@@ -342,7 +382,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_sub_align_s16 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(short __unaligned *)p -= val;");
+		output_storage ("	*(short " UNALIGNED_ATTRIBUTE "*)p -= val;");
 		output_storage ("}");
 		return;
 
@@ -350,7 +390,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_sub_align_u32 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(unsigned int __unaligned *)p -= val;");
+		output_storage ("	*(unsigned int " UNALIGNED_ATTRIBUTE "*)p -= val;");
 		output_storage ("}");
 		return;
 
@@ -358,7 +398,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_sub_align_s32 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(int __unaligned *)p -= val;");
+		output_storage ("	*(int " UNALIGNED_ATTRIBUTE "*)p -= val;");
 		output_storage ("}");
 		return;
 
@@ -366,7 +406,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_sub_align_u64 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(cob_u64_t __unaligned *)p -= val;");
+		output_storage ("	*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p -= val;");
 		output_storage ("}");
 		return;
 
@@ -374,7 +414,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("static COB_INLINE COB_A_INLINE void");
 		output_storage ("cob_sub_align_s64 (void *p, const int val)");
 		output_storage ("{");
-		output_storage ("	*(cob_s64_t __unaligned *)p -= val;");
+		output_storage ("	*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p -= val;");
 		output_storage ("}");
 		return;
 
@@ -387,7 +427,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	if (unlikely(n < 0)) {");
 		output_storage ("		return 1;");
 		output_storage ("	}");
-		output_storage ("	val = COB_BSWAP_16 (*(unsigned short __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_16 (*(unsigned short " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -398,7 +438,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("{");
 		output_storage ("	short	val;");
 
-		output_storage ("	val = COB_BSWAP_16 (*(short __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_16 (*(short " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -412,7 +452,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	if (unlikely(n < 0)) {");
 		output_storage ("		return 1;");
 		output_storage ("	}");
-		output_storage ("	val = COB_BSWAP_32 (*(unsigned int __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_32 (*(unsigned int " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -423,7 +463,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("{");
 		output_storage ("	int	val;");
 
-		output_storage ("	val = COB_BSWAP_32 (*(int __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_32 (*(int " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -437,7 +477,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	if (unlikely(n < 0)) {");
 		output_storage ("		return 1;");
 		output_storage ("	}");
-		output_storage ("	val = COB_BSWAP_64 (*(cob_u64_t __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_64 (*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -448,7 +488,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("{");
 		output_storage ("	cob_s64_t	val;");
 
-		output_storage ("	val = COB_BSWAP_64 (*(cob_s64_t __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_64 (*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	return (val < n) ? -1 : (val > n);");
 		output_storage ("}");
 		return;
@@ -487,7 +527,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("		return 1;");
 		output_storage ("	}");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = *(const unsigned short __unaligned *)p;");
+		output_storage ("	val = *(const unsigned short " UNALIGNED_ATTRIBUTE "*)p;");
 #else
 		output_storage ("	x = &val;");
 		output_storage ("	optim_memcpy (x, p, 2);");
@@ -503,7 +543,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	short	val;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = *(const short __unaligned *)p;");
+		output_storage ("	val = *(const short " UNALIGNED_ATTRIBUTE "*)p;");
 #else
 		output_storage ("	void	*x;");
 
@@ -565,7 +605,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("		return 1;");
 		output_storage ("	}");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = *(const unsigned int __unaligned *)p;");
+		output_storage ("	val = *(const unsigned int " UNALIGNED_ATTRIBUTE "*)p;");
 #else
 		output_storage ("	x = &val;");
 		output_storage ("	optim_memcpy (x, p, 4);");
@@ -581,7 +621,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	int	val;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = *(const int __unaligned *)p;");
+		output_storage ("	val = *(const int " UNALIGNED_ATTRIBUTE "*)p;");
 #else
 		output_storage ("	void	*x;");
 
@@ -716,7 +756,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	cob_u64_t	val;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = *(const cob_u64_t __unaligned *)p;");
+		output_storage ("	val = *(const cob_u64_t " UNALIGNED_ATTRIBUTE "*)p;");
 #else
 		output_storage ("	x = &val;");
 		output_storage ("	optim_memcpy (x, p, 8);");
@@ -732,7 +772,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	cob_s64_t	val;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = *(const cob_s64_t __unaligned *)p;");
+		output_storage ("	val = *(const cob_s64_t " UNALIGNED_ATTRIBUTE "*)p;");
 #else
 		output_storage ("	void		*x;");
 
@@ -766,7 +806,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_add_u16 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(unsigned short __unaligned *)p += val;");
+		output_storage ("	*(unsigned short " UNALIGNED_ATTRIBUTE "*)p += val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	unsigned short	n;");
@@ -784,7 +824,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_add_s16 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(short __unaligned *)p += val;");
+		output_storage ("	*(short " UNALIGNED_ATTRIBUTE "*)p += val;");
 #else
 		output_storage ("	void	*x;");
 		output_storage ("	short	n;");
@@ -844,7 +884,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_add_u32 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(unsigned int __unaligned *)p += val;");
+		output_storage ("	*(unsigned int " UNALIGNED_ATTRIBUTE "*)p += val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	unsigned int	n;");
@@ -862,7 +902,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_add_s32 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(int __unaligned *)p += val;");
+		output_storage ("	*(int " UNALIGNED_ATTRIBUTE "*)p += val;");
 #else
 		output_storage ("	void	*x;");
 		output_storage ("	int	n;");
@@ -1006,7 +1046,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_add_u64 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(cob_u64_t __unaligned *)p += val;");
+		output_storage ("	*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p += val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	cob_u64_t	n;");
@@ -1024,7 +1064,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_add_s64 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(cob_s64_t __unaligned *)p += val;");
+		output_storage ("	*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p += val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	cob_s64_t	n;");
@@ -1058,7 +1098,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_sub_u16 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(unsigned short __unaligned *)p -= val;");
+		output_storage ("	*(unsigned short " UNALIGNED_ATTRIBUTE "*)p -= val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	unsigned short	n;");
@@ -1076,7 +1116,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_sub_s16 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(short __unaligned *)p -= val;");
+		output_storage ("	*(short " UNALIGNED_ATTRIBUTE "*)p -= val;");
 #else
 		output_storage ("	void	*x;");
 		output_storage ("	short	n;");
@@ -1136,7 +1176,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_sub_u32 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(unsigned int __unaligned *)p -= val;");
+		output_storage ("	*(unsigned int " UNALIGNED_ATTRIBUTE "*)p -= val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	unsigned int	n;");
@@ -1154,7 +1194,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_sub_s32 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(int __unaligned *)p -= val;");
+		output_storage ("	*(int " UNALIGNED_ATTRIBUTE "*)p -= val;");
 #else
 		output_storage ("	void	*x;");
 		output_storage ("	int	n;");
@@ -1298,7 +1338,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_sub_u64 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(cob_u64_t __unaligned *)p -= val;");
+		output_storage ("	*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p -= val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	cob_u64_t	n;");
@@ -1316,7 +1356,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("cob_sub_s64 (void *p, const int val)");
 		output_storage ("{");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	*(cob_s64_t __unaligned *)p -= val;");
+		output_storage ("	*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p -= val;");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	cob_s64_t	n;");
@@ -1344,7 +1384,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("		return 1;");
 		output_storage ("	}");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = COB_BSWAP_16 (*(unsigned short __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_16 (*(unsigned short " UNALIGNED_ATTRIBUTE "*)p);");
 #else
 		output_storage ("	x = &val;");
 		output_storage ("	optim_memcpy (x, p, 2);");
@@ -1361,7 +1401,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	short	val;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = COB_BSWAP_16 (*(short __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_16 (*(short " UNALIGNED_ATTRIBUTE "*)p);");
 #else
 		output_storage ("	void	*x;");
 
@@ -1418,7 +1458,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("		return 1;");
 		output_storage ("	}");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = COB_BSWAP_32 (*(const unsigned int __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_32 (*(const unsigned int " UNALIGNED_ATTRIBUTE "*)p);");
 #else
 		output_storage ("	x = &val;");
 		output_storage ("	optim_memcpy (x, p, 4);");
@@ -1435,7 +1475,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	int	val;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = COB_BSWAP_32 (*(const int __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_32 (*(const int " UNALIGNED_ATTRIBUTE "*)p);");
 #else
 		output_storage ("	void	*x;");
 
@@ -1556,7 +1596,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("		return 1;");
 		output_storage ("	}");
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = COB_BSWAP_64 (*(const cob_u64_t __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_64 (*(const cob_u64_t " UNALIGNED_ATTRIBUTE "*)p);");
 #else
 		output_storage ("	x = &val;");
 		output_storage ("	optim_memcpy (x, p, 8);");
@@ -1573,7 +1613,7 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	cob_s64_t	val;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	val = COB_BSWAP_64 (*(const cob_s64_t __unaligned *)p);");
+		output_storage ("	val = COB_BSWAP_64 (*(const cob_s64_t " UNALIGNED_ATTRIBUTE "*)p);");
 #else
 		output_storage ("	void		*x;");
 		output_storage ("	x = &val;");
@@ -1593,9 +1633,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	unsigned short	n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_16 (*(unsigned short __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_16 (*(unsigned short " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n += val;");
-		output_storage ("	*(unsigned short __unaligned *)p = COB_BSWAP_16(n);");
+		output_storage ("	*(unsigned short " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_16(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -1617,9 +1657,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	short		n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_16 (*(short __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_16 (*(short " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n += val;");
-		output_storage ("	*(short __unaligned *)p = COB_BSWAP_16(n);");
+		output_storage ("	*(short " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_16(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -1681,9 +1721,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	unsigned int	n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_32 (*(unsigned int __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_32 (*(unsigned int " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n += val;");
-		output_storage ("	*(unsigned int __unaligned *)p = COB_BSWAP_32(n);");
+		output_storage ("	*(unsigned int " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_32(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -1709,9 +1749,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	int		n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_32 (*(int __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_32 (*(int " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n += val;");
-		output_storage ("	*(int __unaligned *)p = COB_BSWAP_32(n);");
+		output_storage ("	*(int " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_32(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -1893,9 +1933,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	cob_u64_t		n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_64 (*(cob_u64_t __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_64 (*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n += val;");
-		output_storage ("	*(cob_u64_t __unaligned *)p = COB_BSWAP_64(n);");
+		output_storage ("	*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_64(n);");
 #else
 		output_storage ("	unsigned char		*x;");
 		output_storage ("	unsigned char		*px = p;");
@@ -1929,9 +1969,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	cob_s64_t		n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_64 (*(cob_s64_t __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_64 (*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n += val;");
-		output_storage ("	*(cob_s64_t __unaligned *)p = COB_BSWAP_64(n);");
+		output_storage ("	*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_64(n);");
 #else
 		output_storage ("	unsigned char		*x;");
 		output_storage ("	unsigned char		*px = p;");
@@ -1967,9 +2007,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	unsigned short	n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_16 (*(unsigned short __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_16 (*(unsigned short " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n -= val;");
-		output_storage ("	*(unsigned short __unaligned *)p = COB_BSWAP_16(n);");
+		output_storage ("	*(unsigned short " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_16(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -1991,9 +2031,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	short		n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_16 (*(short __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_16 (*(short " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n -= val;");
-		output_storage ("	*(short __unaligned *)p = COB_BSWAP_16(n);");
+		output_storage ("	*(short " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_16(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -2055,9 +2095,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	unsigned int	n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_32 (*(unsigned int __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_32 (*(unsigned int " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n -= val;");
-		output_storage ("	*(unsigned int __unaligned *)p = COB_BSWAP_32(n);");
+		output_storage ("	*(unsigned int " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_32(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -2083,9 +2123,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	int		n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_32 (*(int __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_32 (*(int " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n -= val;");
-		output_storage ("	*(int __unaligned *)p = COB_BSWAP_32(n);");
+		output_storage ("	*(int " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_32(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -2267,9 +2307,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	cob_u64_t	n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_64 (*(cob_u64_t __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_64 (*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n -= val;");
-		output_storage ("	*(cob_u64_t __unaligned *)p = COB_BSWAP_64(n);");
+		output_storage ("	*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_64(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -2303,9 +2343,9 @@ cob_gen_optim (const enum cb_optim val)
 		output_storage ("	cob_s64_t		n;");
 
 #ifdef	COB_ALLOW_UNALIGNED
-		output_storage ("	n = COB_BSWAP_64 (*(cob_s64_t __unaligned *)p);");
+		output_storage ("	n = COB_BSWAP_64 (*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p);");
 		output_storage ("	n -= val;");
-		output_storage ("	*(cob_s64_t __unaligned *)p = COB_BSWAP_64(n);");
+		output_storage ("	*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_64(n);");
 #else
 		output_storage ("	unsigned char		*x;");
 		output_storage ("	unsigned char		*px = p;");
@@ -2341,7 +2381,7 @@ cob_gen_optim (const enum cb_optim val)
 
 #ifdef	COB_ALLOW_UNALIGNED
 		output_storage ("	n = val;");
-		output_storage ("	*(unsigned short __unaligned *)p = COB_BSWAP_16(n);");
+		output_storage ("	*(unsigned short " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_16(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -2362,7 +2402,7 @@ cob_gen_optim (const enum cb_optim val)
 
 #ifdef	COB_ALLOW_UNALIGNED
 		output_storage ("	n = val;");
-		output_storage ("	*(short __unaligned *)p = COB_BSWAP_16(n);");
+		output_storage ("	*(short " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_16(n);");
 #else
 		output_storage ("	unsigned char	*x;");
 		output_storage ("	unsigned char	*px = p;");
@@ -2415,7 +2455,7 @@ cob_gen_optim (const enum cb_optim val)
 
 #ifdef	COB_ALLOW_UNALIGNED
 		output_storage ("	n = val;");
-		output_storage ("	*(unsigned int __unaligned *)p = COB_BSWAP_32(n);");
+		output_storage ("	*(unsigned int " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_32(n);");
 #else
 		output_storage ("	unsigned char		*x;");
 		output_storage ("	unsigned char		*px = p;");
@@ -2438,7 +2478,7 @@ cob_gen_optim (const enum cb_optim val)
 
 #ifdef	COB_ALLOW_UNALIGNED
 		output_storage ("	n = val;");
-		output_storage ("	*(int __unaligned *)p = COB_BSWAP_32(n);");
+		output_storage ("	*(int " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_32(n);");
 #else
 		output_storage ("	unsigned char		*x;");
 		output_storage ("	unsigned char		*px = p;");
@@ -2575,7 +2615,7 @@ cob_gen_optim (const enum cb_optim val)
 
 #ifdef	COB_ALLOW_UNALIGNED
 		output_storage ("	n = val;");
-		output_storage ("	*(cob_u64_t __unaligned *)p = COB_BSWAP_64(n);");
+		output_storage ("	*(cob_u64_t " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_64(n);");
 #else
 		output_storage ("	unsigned char		*x;");
 		output_storage ("	unsigned char		*px = p;");
@@ -2602,7 +2642,7 @@ cob_gen_optim (const enum cb_optim val)
 
 #ifdef	COB_ALLOW_UNALIGNED
 		output_storage ("	n = val;");
-		output_storage ("	*(cob_s64_t __unaligned *)p = COB_BSWAP_64(n);");
+		output_storage ("	*(cob_s64_t " UNALIGNED_ATTRIBUTE "*)p = COB_BSWAP_64(n);");
 #else
 		output_storage ("	unsigned char		*x;");
 		output_storage ("	unsigned char		*px = p;");
@@ -2620,10 +2660,10 @@ cob_gen_optim (const enum cb_optim val)
 #endif
 		output_storage ("}");
 		return;
+	/* LCOV_EXCL_START */
 	default:
-		/* LCOV_EXCL_START */
 		cobc_err_msg (_("unexpected optimization value: %d"), val);
 		COBC_ABORT ();
-		/* LCOV_EXCL_STOP */
+	/* LCOV_EXCL_STOP */
 	}
 }
