@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2003-2012, 2014-2017 Free Software Foundation, Inc.
+   Copyright (C) 2003-2012, 2014-2020 Free Software Foundation, Inc.
    Written by Keisuke Nishida, Roger While, Simon Sobisch, Ron Norman
 
    This file is part of GnuCOBOL.
@@ -15,12 +15,11 @@
    GNU Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public License
-   along with GnuCOBOL.  If not, see <http://www.gnu.org/licenses/>.
+   along with GnuCOBOL.  If not, see <https://www.gnu.org/licenses/>.
 */
 
 
-#include "config.h"
-#include "defaults.h"
+#include <config.h>
 
 #ifndef	_GNU_SOURCE
 #define _GNU_SOURCE	1
@@ -38,6 +37,15 @@
 #include <unistd.h>
 #endif
 
+#ifdef HAVE_FMEMOPEN
+#if defined(HAVE_DECL_FMEMOPEN) && HAVE_DECL_FMEMOPEN == 0
+/* function available and working, declaration missing on AIX... */
+FILE *fmemopen (void *buf, size_t size, const char *mode);
+#endif
+#endif
+
+#include <errno.h>
+
 /*	NOTE - The following variable should be uncommented when
 	it is known that dlopen(NULL) is borked.
 	This is known to be true for some PA-RISC HP-UX 11.11 systems.
@@ -50,6 +58,7 @@
 
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
+#include <io.h>	/* for access */
 
 static HMODULE
 lt_dlopen (const char *x)
@@ -72,10 +81,6 @@ lt_dlsym (HMODULE hmod, const char *p)
 	return modun.voidptr;
 }
 
-#if	0	/* RXWRXW - Win dlsym */
-#define lt_dlsym(x,y)	GetProcAddress(x, y)
-#endif
-
 #define lt_dlclose(x)	FreeLibrary(x)
 #define	lt_dlinit()
 #define	lt_dlexit()
@@ -111,7 +116,6 @@ lt_dlerror (void)
 
 /* Force symbol exports */
 #define	COB_LIB_EXPIMP
-
 #include "libcob.h"
 #include "coblocal.h"
 
@@ -167,14 +171,16 @@ static char			*resolve_alloc;
 static char			*resolve_error_buff;
 static void			*call_buffer;
 static char			*call_filename_buff;
-static char			*call_entry_buff;
-static unsigned char		*call_entry2_buff;
 
+#ifndef	COB_BORKED_DLOPEN
 static lt_dlhandle		mainhandle;
+#endif
 
 static size_t			call_lastsize;
-static size_t			resolve_size;
+static size_t			resolve_size = 0;
 static unsigned int		cob_jmp_primed;
+static cob_field_attr	const_float_attr =
+			{COB_TYPE_NUMERIC_DOUBLE, 8, 0, COB_FLAG_HAVE_SIGN, NULL};
 static cob_field_attr	const_binll_attr =
 			{COB_TYPE_NUMERIC_BINARY, 18, 0, COB_FLAG_HAVE_SIGN, NULL};
 static cob_field_attr	const_binull_attr =
@@ -259,6 +265,7 @@ static const unsigned char	valid_char[256] = {
 	['z'] = 1
 };
 #else
+static int					init_valid_char = 1;
 static unsigned char		valid_char[256];
 static const unsigned char	pvalid_char[] =
 	"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz";
@@ -273,13 +280,29 @@ set_resolve_error (void)
 	cob_set_exception (COB_EC_PROGRAM_NOT_FOUND);
 }
 
-static void
-cob_set_library_path (const char *path)
+static int last_entry_is_working_directory (const char *buff, const char *pstr)
 {
+	const size_t pos = pstr - buff;	/* always > 2 */
+	if (buff[pos - 1] == '.'
+	 && buff[pos - 2] == PATHSEP_CHAR) {
+		return 1;
+	}
+	return 0;
+}
+
+/* resolves the actual library path used from
+   * COB_LIBRARY_PATH runtime setting
+   * "." as current working direktory [if not included already: prefixed]
+   * COB_LIBRARY_PATH inbuilt (which normally includes modules
+     like CBL_OC_DUMP) [if not included already: appended]
+*/
+static void
+cob_set_library_path ()
+{
+	char		buff[COB_MEDIUM_BUFF];
 	char		*p;
 	char		*pstr;
 	size_t		i;
-	size_t		size;
 	struct stat	st;
 
 	int 		flag;
@@ -290,59 +313,109 @@ cob_set_library_path (const char *path)
 		cob_free (resolve_alloc);
 	}
 
-	/* Count the number of separators */
+	/* setup buffer and count number of separators,
+	   check for "." */
 	i = 1;
-	size = 0;
-	for (p = (char *)path; *p; p++, size++) {
-		if (*p == PATHSEP_CHAR) {
-			i++;
+	pstr = buff + 2; /* leaving place for prefixing the working directory */
+	buff[0] = ' ';
+	buff[1] = PATHSEP_CHAR;
+	flag = 0;
+	if (cobsetptr->cob_library_path != NULL
+	 && strcmp (cobsetptr->cob_library_path, ".") != 0) {
+		for (p = cobsetptr->cob_library_path; *p; p++, pstr++) {
+#ifdef	_WIN32
+			if (*p == '/') {
+				*pstr = '\\';
+				continue;
+			}
+#else
+			if (*p == '\\') {
+				*pstr = '/';
+				continue;
+			}
+#endif
+			if (*p == PATHSEP_CHAR) {
+				i++;
+				flag |= last_entry_is_working_directory (buff, pstr);
+			}
+			*pstr = *p;
 		}
+		*pstr = PATHSEP_CHAR;
+		i++;
+		flag |= last_entry_is_working_directory (buff, pstr);
+		pstr++;
 	}
 
-	/* Build path array */
-	size++;
-	resolve_alloc = cob_malloc (size);
-	pstr = resolve_alloc;
-	for (p = (char *)path; *p; p++, pstr++) {
+#ifdef COB_LIBRARY_PATH
+	if (COB_LIBRARY_PATH[0] != 0
+	 && strcmp (COB_LIBRARY_PATH, ".") != 0) {
+		for (p = (char *)COB_LIBRARY_PATH; *p; p++, pstr++) {
 #ifdef	_WIN32
-		if (*p == (unsigned char)'/') {
-			*pstr = (unsigned char)'\\';
-			continue;
-		}
+			if (*p == '/') {
+				*pstr = '\\';
+				continue;
+			}
 #else
-		if (*p == (unsigned char)'\\') {
-			*pstr = (unsigned char)'/';
-			continue;
-		}
+			if (*p == '\\') {
+				*pstr = '/';
+				continue;
+			}
 #endif
-		*pstr = *p;
+			if (*p == PATHSEP_CHAR) {
+				i++;
+			}
+			*pstr = *p;
+		}
 	}
+#endif
+	*pstr = 0;
+
+	/* prefix working directory if missing */
+	if (!flag) {
+		buff[0] = '.';
+		i++;
+		p = buff;
+	} else {
+		p = buff + 2;
+	}
+	/* Build path array */
+	resolve_alloc = cob_strdup (p);
+	pstr = resolve_alloc;
 
 	resolve_path = cob_malloc (sizeof (char *) * i);
 	resolve_size = 0;
-	pstr = resolve_alloc;
+
 	for (; ; ) {
 		p = strtok (pstr, PATHSEP_STR);
 		if (!p) {
 			break;
 		}
 		pstr = NULL;
+
+		/* check if directory
+		   (note: entries like X:\ _must_ be specified with trailing slash !) */
 		if (stat (p, &st) || !(S_ISDIR (st.st_mode))) {
+			/* possibly raise a warning, maybe only if explicit asked */
 			continue;
 		}
 
-		/*
-		 * look if we already have this path
-		 */
+		/* remove trailing slash from entry (always added on use) */
+		i = strlen (p) - 1;
+		if (p[i] == SLASH_CHAR) {
+			p[i] = 0;
+		}
+
+		/* check if we have this path already */
 		flag = 0;
 		for (i = 0; i < resolve_size; i++) {
-			if(strcmp(resolve_path[i], p) == 0) {
+			if (strcmp (resolve_path[i], p) == 0) {
 				flag = 1;
 				break;
 			}
 		}
 
-		if (!flag) {
+		/* finally: new entry for the resolve path */
+		if (flag == 0) {
 			resolve_path[resolve_size++] = p;
 		}
 	}
@@ -357,6 +430,10 @@ do_cancel_module (struct call_hash *p, struct call_hash **base_hash,
 	int nocancel;
 	nocancel = 0;
 
+	/* FIXME: check the modules entry point and take care of
+	   CBL_ERROR_PROC / CBL_EXIT_PROC which defines: If a program containing
+	   an exit/error procedure is canceled, the exit/error procedure is removed. */
+
 	if (!p->module) {
 		return;
 	}
@@ -377,7 +454,7 @@ do_cancel_module (struct call_hash *p, struct call_hash **base_hash,
 		nocancel = 1;
 	}
 #ifdef _MSC_VER
-#pragma warning(suppress: 4113) // funcint is a generic function prototype
+#pragma warning(suppress: 4113) /* funcint is a generic function prototype */
 	cancel_func = p->module->module_cancel.funcint;
 #else
 	cancel_func = p->module->module_cancel.funcint;
@@ -548,10 +625,7 @@ insert (const char *name, void *func, lt_dlhandle handle,
 	p->handle = handle;
 	p->module = module;
 	if (path) {
-#ifdef	_WIN32
-		/* Malloced path or NULL */
-		p->path = _fullpath (NULL, path, 1);
-#elif	defined(HAVE_CANONICALIZE_FILE_NAME)
+#if	defined(HAVE_CANONICALIZE_FILE_NAME)
 		/* Malloced path or NULL */
 		p->path = canonicalize_file_name (path);
 #elif	defined(HAVE_REALPATH)
@@ -562,6 +636,9 @@ insert (const char *name, void *func, lt_dlhandle handle,
 			p->path = cob_strdup (s);
 		}
 		cob_free (s);
+#elif	defined	(_WIN32)
+		/* Malloced path or NULL */
+		p->path = _fullpath (NULL, path, 1);
 #endif
 		if (!p->path) {
 			p->path = cob_strdup (path);
@@ -596,23 +673,118 @@ lookup (const char *name)
 	return NULL;
 }
 
+static int
+cob_encode_invalid_chars (const unsigned char* const name,
+	unsigned char* const name_buff,
+	const int buff_size, int *external_pos)
+{
+	const unsigned char *s = name;
+	int pos = *external_pos;
+
+#ifndef	HAVE_DESIGNATED_INITS
+	if (init_valid_char) {
+		const unsigned char	*pv;
+		init_valid_char = 0;
+		memset (valid_char, 0, sizeof(valid_char));
+		for (pv = pvalid_char; *pv; ++pv) {
+			valid_char[*pv] = 1;
+		}
+	}
+#endif
+
+	/* Encode invalid letters */
+	for (; *s; ++s) {
+		if (pos >= buff_size - 3) {
+			name_buff[pos] = 0;
+			return -pos;
+		}
+		if (likely (valid_char[*s])) {
+			name_buff[pos++] = *s;
+		} else {
+			name_buff[pos++] = (unsigned char)'_';
+			if (*s == (unsigned char)'-') {
+				name_buff[pos++] = (unsigned char)'_';
+			} else {
+				name_buff[pos++] = hexval[*s / 16U];
+				name_buff[pos++] = hexval[*s % 16U];
+			}
+		}
+	}
+
+	*external_pos = pos;
+	return pos;
+}
+
+/** encode given name
+  \param name to encode
+  \param name_buff to place the encoded name to
+  \param buff_size available
+  \param fold_case may be COB_FOLD_UPPER or COB_FOLD_LOWER
+  \return size of the encoded name, negative if the buffer size would be exceeded
+ */
+int
+cob_encode_program_id (const unsigned char *const name,
+	unsigned char *const name_buff,
+	const int buff_size, const int fold_case)
+{
+	int pos = 0;
+	/* Encode the initial digit */
+	if (unlikely (*name <= (unsigned char)'9' && *name >= (unsigned char)'0')) {
+		name_buff[pos++] = (unsigned char)'_';
+	}
+	/* Encode invalid letters */
+	cob_encode_invalid_chars (name, name_buff, buff_size, &pos);
+
+	name_buff[pos] = 0;
+
+	/* Check case folding */
+	switch (fold_case) {
+	case COB_FOLD_NONE:
+		break;
+	case COB_FOLD_UPPER:
+	{
+		unsigned char *p;
+		for (p = name_buff; *p; p++) {
+			if (islower (*p)) {
+				*p = (cob_u8_t)toupper (*p);
+			}
+		}
+		break;
+	}
+	case COB_FOLD_LOWER:
+	{
+		unsigned char *p;
+		for (p = name_buff; *p; p++) {
+			if (isupper (*p)) {
+				*p = (cob_u8_t)tolower (*p);
+			}
+		}
+		break;
+	}
+	default:
+		break;
+	}
+	
+	return pos;
+}
+
 static void *
 cob_resolve_internal (const char *name, const char *dirent,
-		      const int fold_case)
+	const int fold_case)
 {
-	unsigned char		*p;
 	const unsigned char	*s;
 	void			*func;
 	struct struct_handle	*preptr;
 	lt_dlhandle		handle;
 	size_t			i;
+	char call_entry_buff[COB_MINI_BUFF];
 
 	/* LCOV_EXCL_START */
 	if (unlikely(!cobglobptr)) {
 		cob_fatal_error (COB_FERROR_INITIALIZED);
 	}
 	/* LCOV_EXCL_STOP */
-	cob_set_exception (0);
+	cobglobptr->cob_exception_code = 0;
 
 	/* Search the cache */
 	func = lookup (name);
@@ -620,47 +792,13 @@ cob_resolve_internal (const char *name, const char *dirent,
 		return func;
 	}
 
-	/* Encode program name */
-	p = (unsigned char *)call_entry_buff;
 	s = (const unsigned char *)name;
-	if (unlikely(*s <= (unsigned char)'9' && *s >= (unsigned char)'0')) {
-		*p++ = (unsigned char)'_';
-	}
-	for (; *s; ++s) {
-		if (likely(valid_char[*s])) {
-			*p++ = *s;
-		} else {
-			*p++ = (unsigned char)'_';
-			if (*s == (unsigned char)'-') {
-				*p++ = (unsigned char)'_';
-			} else {
-				*p++ = hexval[*s / 16U];
-				*p++ = hexval[*s % 16U];
-			}
-		}
-	}
-	*p = 0;
 
-	/* Check case folding */
-	switch (fold_case) {
-	case COB_FOLD_UPPER:
-		for (p = (unsigned char *)call_entry_buff; *p; p++) {
-			if (islower (*p)) {
-				*p = (cob_u8_t)toupper (*p);
-			}
-		}
-		break;
-	case COB_FOLD_LOWER:
-		for (p = (unsigned char *)call_entry_buff; *p; p++) {
-			if (isupper (*p)) {
-				*p = (cob_u8_t)tolower (*p);
-			}
-		}
-		break;
-	default:
-		break;
-	}
+	/* Encode program name, including case folding */
+	cob_encode_program_id (s, (unsigned char *)call_entry_buff,
+		COB_MINI_MAX, fold_case);
 
+#ifndef	COB_BORKED_DLOPEN
 	/* Search the main program */
 	if (mainhandle != NULL) {
 		func = lt_dlsym (mainhandle, call_entry_buff);
@@ -670,6 +808,7 @@ cob_resolve_internal (const char *name, const char *dirent,
 			return func;
 		}
 	}
+#endif
 
 	/* Search preloaded modules */
 	for (preptr = base_preload_ptr; preptr; preptr = preptr->next) {
@@ -710,10 +849,8 @@ cob_resolve_internal (const char *name, const char *dirent,
 
 	/* Check if name needs conversion */
 	if (unlikely(cobsetptr->name_convert != 0)) {
-		if (!call_entry2_buff) {
-			call_entry2_buff = cob_malloc ((size_t)COB_SMALL_BUFF);
-		}
-		p = call_entry2_buff;
+		unsigned char call_entry2_buff[COB_MINI_BUFF];
+		unsigned char *p = call_entry2_buff;
 		for (; *s; ++s, ++p) {
 			if (cobsetptr->name_convert == 1 && isupper (*s)) {
 				*p = (cob_u8_t) tolower (*s);
@@ -724,7 +861,7 @@ cob_resolve_internal (const char *name, const char *dirent,
 			}
 		}
 		*p = 0;
-		s = (const unsigned char *)call_entry2_buff;
+		s = call_entry2_buff;
 	}
 
 	/* Search external modules */
@@ -885,7 +1022,7 @@ void
 cob_call_error (void)
 {
 	cob_runtime_error ("%s", cob_resolve_error ());
-	cob_stop_run (1);
+	cob_stop_run (EXIT_FAILURE);
 }
 
 void
@@ -955,7 +1092,7 @@ cob_resolve_func (const char *name)
 	p = cob_resolve_internal (name, NULL, 0);
 	if (unlikely(!p)) {
 		cob_runtime_error (_("user-defined FUNCTION '%s' not found"), name);
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	return p;
 }
@@ -970,7 +1107,6 @@ cob_call_field (const cob_field *f, const struct cob_call_struct *cs,
 	char				*buff;
 	char				*entry;
 	char				*dirent;
-	int				len;
 
 	/* LCOV_EXCL_START */
 	if (unlikely(!cobglobptr)) {
@@ -983,10 +1119,11 @@ cob_call_field (const cob_field *f, const struct cob_call_struct *cs,
 
 	/* check for uncommon leading space - trim it */
 	if (*buff == ' ') {
+		size_t				len;
 		/* same warning as in cobc/typeck.c */
 		cob_runtime_warning (
 			_("'%s' literal includes leading spaces which are omitted"), buff);
-		len = strlen(buff);
+		len = strlen (buff);
 		while (*buff == ' ') {
 			memmove (buff, buff + 1, --len);
 		}
@@ -1045,7 +1182,7 @@ cob_cancel (const char *name)
 	}
 	if (unlikely(!name)) {
 		cob_runtime_error (_("NULL parameter passed to '%s'"), "cob_cancel");
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	/* LCOV_EXCL_STOP */
 	entry = cob_chk_dirp (name);
@@ -1093,7 +1230,7 @@ cob_cancel_field (const cob_field *f, const struct cob_call_struct *cs)
 		if (!strcmp (entry, s->cob_cstr_name)) {
 			if (s->cob_cstr_cancel.funcvoid) {
 #ifdef _MSC_VER
-#pragma warning(suppress: 4113) // funcint is a generic function prototype
+#pragma warning(suppress: 4113) /* funcint is a generic function prototype */
 				cancel_func = s->cob_cstr_cancel.funcint;
 #else
 				cancel_func = s->cob_cstr_cancel.funcint;
@@ -1120,11 +1257,11 @@ cob_call (const char *name, const int argc, void **argv)
 	}
 	if (argc < 0 || argc > MAX_CALL_FIELD_PARAMS) {
 		cob_runtime_error (_("invalid number of arguments passed to '%s'"), "cob_call");
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	if (unlikely(!name)) {
 		cob_runtime_error (_("NULL parameter passed to '%s'"), "cob_call");
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	/* LCOV_EXCL_STOP */
 	unifunc.funcvoid = cob_resolve_cobol (name, 0, 1);
@@ -1231,6 +1368,7 @@ cob_func (const char *name, const int argc, void **argv)
 	return ret;
 }
 
+#ifndef COB_WITHOUT_JMP
 void *
 cob_savenv (struct cobjmp_buf *jbuf)
 {
@@ -1240,11 +1378,11 @@ cob_savenv (struct cobjmp_buf *jbuf)
 	}
 	if (unlikely(!jbuf)) {
 		cob_runtime_error (_("NULL parameter passed to '%s'"), "cob_savenv");
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	if (cob_jmp_primed) {
 		cob_runtime_error (_("multiple call to 'cob_setjmp'"));
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	/* LCOV_EXCL_STOP */
 	cob_jmp_primed = 1;
@@ -1268,16 +1406,17 @@ cob_longjmp (struct cobjmp_buf *jbuf)
 	}
 	if (unlikely(!jbuf)) {
 		cob_runtime_error (_("NULL parameter passed to '%s'"), "cob_longjmp");
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	if (!cob_jmp_primed) {
 		cob_runtime_error (_("call to 'cob_longjmp' with no prior 'cob_setjmp'"));
-		cob_stop_run (1);
+		cob_stop_run (EXIT_FAILURE);
 	}
 	/* LCOV_EXCL_STOP */
 	cob_jmp_primed = 0;
 	longjmp (jbuf->cbj_jmp_buf, 1);
 }
+#endif
 
 void
 cob_exit_call (void)
@@ -1295,14 +1434,6 @@ cob_exit_call (void)
 		cob_free (call_filename_buff);
 		call_filename_buff = NULL;
 	}
-	if (call_entry_buff) {
-		cob_free (call_entry_buff);
-		call_entry_buff = NULL;
-	}
-	if (call_entry2_buff) {
-		cob_free (call_entry2_buff);
-		call_entry2_buff = NULL;
-	}
 	if (call_buffer) {
 		cob_free (call_buffer);
 		call_buffer = NULL;
@@ -1318,6 +1449,7 @@ cob_exit_call (void)
 	if (resolve_path) {
 		cob_free (resolve_path);
 		resolve_path = NULL;
+		resolve_size = 0;
 	}
 
 #ifndef	COB_ALT_HASH
@@ -1375,17 +1507,20 @@ cob_exit_call (void)
 #if	!defined(_WIN32) && !defined(USE_LIBDL)
 	lt_dlexit ();
 #if	0	/* RXWRXW - ltdl leak */
+#ifndef	COB_BORKED_DLOPEN
 	/* Weird - ltdl leaks mainhandle - This appears to work but .. */
-	cob_free (mainhandle);
+	if (mainhandle) {
+		cob_free (mainhandle);
+	}
+#endif
 #endif
 #endif
 
 }
 
 void
-cob_init_call (cob_global *lptr, cob_settings* sptr)
+cob_init_call (cob_global *lptr, cob_settings* sptr, const int check_mainhandle)
 {
-	char				*buff;
 	char				*s;
 	char				*p;
 	size_t				i;
@@ -1404,18 +1539,12 @@ cob_init_call (cob_global *lptr, cob_settings* sptr)
 	resolve_path = NULL;
 	resolve_alloc = NULL;
 	resolve_error = NULL;
-	resolve_error_buff = NULL;
-	mainhandle = NULL;
 	call_buffer = NULL;
-	call_filename_buff = NULL;
-	call_entry_buff = NULL;
-	call_entry2_buff = NULL;
-	call_table = NULL;
 	call_lastsize = 0;
-	resolve_size = 0;
 	cob_jmp_primed = 0;
 
 #ifndef	HAVE_DESIGNATED_INITS
+	init_valid_char = 0;
 	memset (valid_char, 0, sizeof(valid_char));
 	for (pv = pvalid_char; *pv; ++pv) {
 		valid_char[*pv] = 1;
@@ -1427,37 +1556,28 @@ cob_init_call (cob_global *lptr, cob_settings* sptr)
 
 #ifndef	COB_ALT_HASH
 	call_table = cob_malloc (sizeof (struct call_hash *) * HASH_SIZE);
+#else
+	call_table = NULL;
 #endif
 
-	call_filename_buff = cob_malloc ((size_t)COB_NORMAL_BUFF);
-	call_entry_buff = cob_malloc ((size_t)COB_SMALL_BUFF);
-
-	buff = cob_fast_malloc ((size_t)COB_MEDIUM_BUFF);
-	if (cobsetptr->cob_library_path == NULL
-	 || strcmp(cobsetptr->cob_library_path, ".") == 0) {
-		if (strcmp(COB_LIBRARY_PATH, ".") == 0) {
-			snprintf (buff, (size_t)COB_MEDIUM_MAX, ".");
-		} else {
-			snprintf (buff, (size_t)COB_MEDIUM_MAX, ".%c%s",
-				  PATHSEP_CHAR, COB_LIBRARY_PATH);
-		}
-	} else {
-		if (strcmp(COB_LIBRARY_PATH, ".") == 0) {
-			snprintf (buff, (size_t)COB_MEDIUM_MAX, "%s%c.",
-				  cobsetptr->cob_library_path, PATHSEP_CHAR);
-		} else {
-			snprintf (buff, (size_t)COB_MEDIUM_MAX, "%s%c.%c%s",
-				  cobsetptr->cob_library_path, PATHSEP_CHAR, PATHSEP_CHAR, COB_LIBRARY_PATH);
-		}
-	}
-	cob_set_library_path (buff);
+	/* set static vars resolve_path (data in resolve_alloc) and resolve_size */
+	cob_set_library_path ();
 
 	lt_dlinit ();
 
 #ifndef	COB_BORKED_DLOPEN
-	mainhandle = lt_dlopen (NULL);
+	/* only set main handle if not started by cobcrun as this
+	   saves a check for exported functions in every CALL
+	*/
+	if (check_mainhandle) {
+		mainhandle = lt_dlopen (NULL);
+	} else {
+		mainhandle = NULL;
+	}
 #endif
 
+	call_filename_buff = cob_malloc ((size_t)COB_NORMAL_BUFF);
+
 	if (cobsetptr->cob_preload_str != NULL) {
 
 		p = cob_strdup (cobsetptr->cob_preload_str);
@@ -1467,6 +1587,7 @@ cob_init_call (cob_global *lptr, cob_settings* sptr)
 
 		s = strtok (p, PATHSEP_STR);
 		for (; s; s = strtok (NULL, PATHSEP_STR)) {
+			char		buff[COB_MEDIUM_BUFF];
 #ifdef __OS400__
 			for (t = s; *t; ++t) {
 				*t = toupper (*t);
@@ -1474,7 +1595,6 @@ cob_init_call (cob_global *lptr, cob_settings* sptr)
 			cache_preload (t);
 #else
 			for (i = 0; i < resolve_size; ++i) {
-				buff[COB_MEDIUM_MAX] = 0;
 				snprintf (buff, (size_t)COB_MEDIUM_MAX,
 					  "%s%c%s.%s",
 					  resolve_path[i], SLASH_CHAR, s, COB_MODULE_EXT);
@@ -1482,7 +1602,7 @@ cob_init_call (cob_global *lptr, cob_settings* sptr)
 					break;
 				}
 			}
-			/* If not found, try just using the name */
+			/* If not found, try just using the name as-is */
 			if (i == resolve_size) {
 				(void)cache_preload (s);
 			}
@@ -1490,7 +1610,6 @@ cob_init_call (cob_global *lptr, cob_settings* sptr)
 		}
 		cob_free (p);
 	}
-	cob_free (buff);
 	call_buffer = cob_fast_malloc ((size_t)CALL_BUFF_SIZE);
 	call_lastsize = CALL_BUFF_SIZE;
 }
@@ -1499,22 +1618,26 @@ cob_init_call (cob_global *lptr, cob_settings* sptr)
  * Routines for C interface with COBOL
  */
 
-static cob_field *
+cob_field *
 cob_get_param_field (int n, const char *caller_name)
 {
 	if (cobglobptr == NULL
 	 || COB_MODULE_PTR == NULL) {
-		cob_runtime_warning (_("%s: COBOL runtime is not initialized"), caller_name);
+		/* note: same message in call.c */
+		cob_runtime_warning_external (caller_name, 1,
+			_("cob_init() has not been called"));
 		return NULL;
 	}
 	if (n < 1
 	 || n > cobglobptr->cob_call_params) {
-		cob_runtime_warning (_("%s: param %d is not within range of %d"),
-				     caller_name, n, cobglobptr->cob_call_params);
+		cob_runtime_warning_external (caller_name, 1,
+			_("parameter %d is not within range of %d"),
+			n, cobglobptr->cob_call_params);
 		return NULL;
 	}
 	if (COB_MODULE_PTR->cob_procedure_params[n - 1] == NULL) {
-		cob_runtime_warning (_("%s: param %d is NULL"), caller_name, n);
+		cob_runtime_warning_external (caller_name, 1,
+			_("parameter %d is NULL"), n);
 		return NULL;
 	}
 	return COB_MODULE_PTR->cob_procedure_params[n - 1];
@@ -1526,7 +1649,10 @@ cob_get_num_params (void)
 	if (cobglobptr) {
 		return cobglobptr->cob_call_params;
 	}
-	cob_runtime_warning (_("%s COBOL runtime is not initialized"), "cob_get_num_params");
+
+	/* note: same message in call.c */
+	cob_runtime_warning_external ("cob_get_num_params", 1,
+		_("cob_init() has not been called"));
 	return -1;
 }
 
@@ -1534,78 +1660,63 @@ int
 cob_get_param_type (int n)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_get_param_type");
-
-	if (f == NULL) {
-		return -1;
-	}
-	if (f->attr->type == COB_TYPE_NUMERIC_BINARY) {
-		if (COB_FIELD_REAL_BINARY (f)) {
-			return COB_TYPE_NUMERIC_COMP5;
-		}
-#ifndef WORDS_BIGENDIAN
-		if (!COB_FIELD_BINARY_SWAP(f)) {
-			return COB_TYPE_NUMERIC_COMP5;
-		}
-#endif
-	}
-	return (int)f->attr->type;
+	return cob_get_field_type (f);
 }
 
 int
 cob_get_param_size (int n)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_get_param_size");
-
-	if (f == NULL) {
-		return -1;
-	}
-	return (int)f->size;
+	return cob_get_field_size (f);
 }
 
 int
 cob_get_param_sign (int n)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_get_param_sign");
-	if (f == NULL) {
-		return -1;
-	}
-	if (COB_FIELD_HAVE_SIGN(f)) {
-		return 1;
-	}
-	return 0;
+	return cob_get_field_sign (f);
 }
 
 int
 cob_get_param_scale (int n)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_get_param_scale");
-	if (f == NULL) {
-		return -1;
-	}
-	return (int)f->attr->scale;
+	return cob_get_field_scale (f);
 }
 
 int
 cob_get_param_digits (int n)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_get_param_digits");
-	if (f == NULL) {
-		return -1;
-	}
-	return (int)f->attr->digits;
+	return cob_get_field_digits (f);
 }
 
 int
 cob_get_param_constant (int n)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_get_param_constant");
-	if (f == NULL) {
-		return -1;
-	}
-	if (COB_FIELD_CONSTANT (f)) {
-		return 1;
-	}
-	return 0;
+	return cob_get_field_constant (f);
+}
+
+const char *
+cob_get_param_str (int n, char *buffer, size_t size)
+{
+	cob_field	*f = cob_get_param_field (n, "cob_get_param_str");
+	return cob_get_field_str (f, buffer, size);
+}
+
+const char *
+cob_get_param_str_buffered (int n)
+{
+	cob_field	*f = cob_get_param_field (n, "cob_get_param_str_buffered");
+	return cob_get_field_str_buffered (f);
+}
+
+int
+cob_put_param_str (int n, const char *str)
+{
+	cob_field	*f = cob_get_param_field (n, "cob_put_param_str");
+	return cob_put_field_str (f, str);
 }
 
 void *
@@ -1618,6 +1729,67 @@ cob_get_param_data (int n)
 	return (void*)f->data;
 }
 
+double
+cob_get_dbl_param (int n)
+{
+	void		*cbl_data;
+	double		val;
+	cob_field	temp;
+	cob_field_attr   float_attr;
+	cob_field	*f = cob_get_param_field (n, "cob_get_dbl_param");
+
+	if (f == NULL) {
+		return (double)-1;
+	}
+	cbl_data = f->data;
+
+	switch (f->attr->type) {
+	case COB_TYPE_NUMERIC_FLOAT:
+		return (double)cob_get_comp1 (cbl_data);
+	case COB_TYPE_NUMERIC_DOUBLE:
+		return (double)cob_get_comp2 (cbl_data);
+	default:
+		memcpy(&float_attr, &const_float_attr, sizeof(cob_field_attr));
+		float_attr.scale = f->attr->scale;
+		temp.size = 8;
+		temp.data = (unsigned char *)&val;
+		temp.attr = &float_attr;
+		cob_move (f, &temp);
+		return (double)val;
+	}
+}
+
+void
+cob_put_dbl_param (int n, double val)
+{
+	void		*cbl_data;
+	cob_field	temp;
+	cob_field_attr   float_attr;
+	cob_field	*f = cob_get_param_field (n, "cob_get_dbl_param");
+
+	if (f == NULL) {
+		return;
+	}
+	cbl_data = f->data;
+
+	switch (f->attr->type) {
+	case COB_TYPE_NUMERIC_FLOAT:
+		cob_put_comp1 ((float)val, cbl_data);
+		return;
+	case COB_TYPE_NUMERIC_DOUBLE:
+		cob_put_comp2 (val, cbl_data);
+		return;
+	default:
+		memcpy(&float_attr, &const_float_attr, sizeof(cob_field_attr));
+		float_attr.scale = f->attr->scale;
+		temp.size = 8;
+		temp.data = (unsigned char *)&val;
+		temp.attr = &float_attr;
+		cob_move (&temp, f);
+		return;
+	}
+}
+
 cob_s64_t
 cob_get_s64_param (int n)
 {
@@ -1639,22 +1811,20 @@ cob_get_s64_param (int n)
 		return cob_get_s64_pic9 (cbl_data, size);
 	case COB_TYPE_NUMERIC_BINARY:
 #ifndef WORDS_BIGENDIAN
-		if (COB_FIELD_BINARY_SWAP (f)) {
-			return cob_get_s64_compx (cbl_data, size);
+		if (!COB_FIELD_BINARY_SWAP (f)) {
+			return cob_get_s64_comp5 (cbl_data, size);
 		}
-		return cob_get_s64_comp5 (cbl_data, size);
-#else
-		return cob_get_s64_compx (cbl_data, size);
 #endif
+		return cob_get_s64_compx (cbl_data, size);
 	case COB_TYPE_NUMERIC_PACKED:
 		return cob_get_s64_comp3 (cbl_data, size);
 	case COB_TYPE_NUMERIC_FLOAT:
 		dbl = cob_get_comp1 (cbl_data);
-		val = (cob_s64_t)dbl; // possible data loss is explicit requested
+		val = (cob_s64_t)dbl; /* possible data loss is explicit requested */
 		return val;
 	case COB_TYPE_NUMERIC_DOUBLE:
 		dbl = cob_get_comp2 (cbl_data);
-		val = (cob_s64_t)dbl; // possible data loss is explicit requested
+		val = (cob_s64_t)dbl; /* possible data loss is explicit requested */
 		return val;
 	case COB_TYPE_NUMERIC_EDITED:
 		return cob_get_s64_pic9 (cbl_data, size);
@@ -1687,27 +1857,22 @@ cob_get_u64_param (int n)
 	switch (COB_MODULE_PTR->cob_procedure_params[n - 1]->attr->type) {
 	case COB_TYPE_NUMERIC_DISPLAY:
 		return cob_get_u64_pic9 (cbl_data, size);
-
 	case COB_TYPE_NUMERIC_BINARY:
 #ifndef WORDS_BIGENDIAN
-		if (COB_FIELD_BINARY_SWAP (f)) {
-			return cob_get_u64_compx (cbl_data, size);
+		if (!COB_FIELD_BINARY_SWAP (f)) {
+			return cob_get_u64_comp5 (cbl_data, size);
 		}
-		return cob_get_u64_comp5 (cbl_data, size);
-#else
-		return cob_get_u64_compx (cbl_data, size);
 #endif
-
+		return cob_get_u64_compx (cbl_data, size);
 	case COB_TYPE_NUMERIC_PACKED:
 		return cob_get_u64_comp3 (cbl_data, size);
-
 	case COB_TYPE_NUMERIC_FLOAT:
 		dbl = cob_get_comp1 (cbl_data);
-		val = (cob_u64_t)dbl; // possible data loss is explicit requested
+		val = (cob_u64_t)dbl; /* possible data loss is explicit requested */
 		return val;
 	case COB_TYPE_NUMERIC_DOUBLE:
 		dbl = cob_get_comp2 (cbl_data);
-		val = (cob_u64_t)dbl; // possible data loss is explicit requested
+		val = (cob_u64_t)dbl; /* possible data loss is explicit requested */
 		return val;
 	case COB_TYPE_NUMERIC_EDITED:
 		return cob_get_u64_pic9 (cbl_data, size);
@@ -1722,19 +1887,177 @@ cob_get_u64_param (int n)
 }
 
 char *
-cob_get_picx_param (int n, void *char_field, int char_len)
+cob_get_picx_param (int n, void *char_field, size_t char_len)
 {
-	void		*cbl_data;
-	int		size;
 	cob_field	*f = cob_get_param_field (n, "cob_get_picx_param");
-
 	if (f == NULL) {
 		return NULL;
 	}
+	return cob_get_picx (f->data, f->size, char_field, char_len);
+}
 
-	cbl_data = f->data;
-	size    = f->size;
-	return cob_get_picx (cbl_data, size, char_field, char_len);
+int
+cob_get_field_type (const cob_field *f)
+{
+	if (f == NULL) {
+		return -1;
+	}
+	if (f->attr->type == COB_TYPE_NUMERIC_BINARY) {
+		if (COB_FIELD_REAL_BINARY (f)) {
+			return COB_TYPE_NUMERIC_COMP5;
+		}
+#ifndef WORDS_BIGENDIAN
+		if (!COB_FIELD_BINARY_SWAP (f)) {
+			return COB_TYPE_NUMERIC_COMP5;
+		}
+#endif
+	}
+	return (int)f->attr->type;
+}
+
+int
+cob_get_field_size (const cob_field *f)
+{
+	if (f == NULL) {
+		return -1;
+	}
+	return (int)f->size;
+}
+
+int
+cob_get_field_sign (const cob_field *f)
+{
+	if (f == NULL) {
+		return -1;
+	}
+	return COB_FIELD_HAVE_SIGN (f);
+}
+
+int
+cob_get_field_scale (const cob_field *f)
+{
+	if (f == NULL) {
+		return -1;
+	}
+	return (int)f->attr->scale;
+}
+
+int
+cob_get_field_digits (const cob_field *f)
+{
+	if (f == NULL) {
+		return -1;
+	}
+	return (int)f->attr->digits;
+}
+
+int
+cob_get_field_constant (const cob_field *f)
+{
+	if (f == NULL) {
+		return -1;
+	}
+	return COB_FIELD_CONSTANT (f);
+}
+
+const char *
+cob_get_field_str (const cob_field *f, char *buffer, size_t size)
+{
+	if (unlikely (f == NULL)) {
+		return _("NULL field");
+	}
+	/* variable field's and empty literals may be of zero size */
+	if (unlikely (f->size == 0)) {
+		return "";
+	}
+	/* check if field has data assigned (may be a BASED / LINKAGE item) */
+	if (unlikely (f->data == NULL)) {
+		return _("field not allocated");
+	}
+	if (!buffer || !size) {
+		cob_runtime_warning_external ("cob_get_field_str", 0, "bad buffer/size");
+		return "";
+	}
+	{
+		FILE *fp;
+#ifdef HAVE_FMEMOPEN
+		fp = fmemopen (buffer, size, "w");
+#else
+		fp = cob_create_tmpfile ("display");
+#endif
+		if (fp) {
+			unsigned char pretty = COB_MODULE_PTR->flag_pretty_display;
+			COB_MODULE_PTR->flag_pretty_display = 1;
+			cob_display_common (f, fp);
+#ifndef HAVE_FMEMOPEN
+			{
+				int cur_pos = ftell (fp);
+				if (cur_pos >= 0) {
+					size_t	pos = (size_t) cur_pos;
+					fseek (fp, 0, SEEK_SET);
+					fread ((void*)buffer, 1, pos, fp);
+					if (size > pos) buffer[pos] = 0;
+				}
+			}
+#endif
+			fclose (fp);
+			COB_MODULE_PTR->flag_pretty_display = pretty;
+		}
+	}
+	return buffer;
+}
+
+const char *
+cob_get_field_str_buffered (const cob_field *f)
+{
+	char	*buff = NULL;
+	size_t	size = cob_get_field_size (f) + 1;
+
+	if (size > 0) {
+		if (size < 32) {
+			size = 32;
+		}
+		buff = cob_get_buff (size);
+	}
+	return cob_get_field_str (f, buff, size);
+}
+
+int
+cob_put_field_str (const cob_field *dst, const char *str)
+{
+	const cob_field_attr	const_alpha_attr =
+			{COB_TYPE_ALPHANUMERIC, 0, 0, 0, NULL};
+	cob_field wrk;
+
+	if (!dst ||!str) return EINVAL;
+
+	/* come back later for DYNAMIC LENGTH fields */
+	if (dst->size <= 0) return EINVAL;
+
+	if (COB_FIELD_CONSTANT (dst)) {
+		cob_runtime_warning_external ("cob_put_field_str", 0,
+			_ ("attempt to over-write constant field with '%s'"),
+			str);
+		return EINVAL;
+	}
+
+
+	wrk.attr = &const_alpha_attr;
+	wrk.size = strlen (str);
+	wrk.data = (unsigned char *)str;
+
+	if (COB_FIELD_IS_NUMERIC (dst)) {
+		if (COB_FIELD_TYPE (dst) & COB_TYPE_NUMERIC_FLOAT
+		 || COB_FIELD_TYPE (dst) & COB_TYPE_NUMERIC_DOUBLE) {
+			if (cob_check_numval_f (&wrk)) return 1;
+			wrk = *cob_intr_numval_f (&wrk);
+		} else {
+			if (cob_check_numval (&wrk, NULL, 0, 1)) return 1;
+			wrk = *cob_intr_numval (&wrk);
+		}
+	}
+	cob_move (&wrk, (cob_field *)dst);
+	return 0;
 }
 
 void
@@ -1751,42 +2074,36 @@ cob_put_s64_param (int n, cob_s64_t val)
 		return;
 	}
 
-	cbl_data = f->data;
-	size = f->size;
 	if (COB_FIELD_CONSTANT (f)) {
-		cob_runtime_warning (_("%s: attempt to over-write constant param %d with " CB_FMT_LLD),
-						"cob_put_s64_param", n, val);
+		cob_runtime_warning_external ("cob_put_s64_param", 1,
+			_("attempt to over-write constant parameter %d with " CB_FMT_LLD),
+			n, val);
 		return;
 	}
-
+	cbl_data = f->data;
+	size = f->size;
 	switch (f->attr->type) {
 	case COB_TYPE_NUMERIC_DISPLAY:
 		cob_put_s64_pic9 (val, cbl_data, size);
 		return;
-
 	case COB_TYPE_NUMERIC_BINARY:
 #ifndef WORDS_BIGENDIAN
-		if (COB_FIELD_BINARY_SWAP (f)) {
-			cob_put_s64_compx (val, cbl_data, size);
-		} else {
+		if (!COB_FIELD_BINARY_SWAP (f)) {
 			cob_put_s64_comp5 (val, cbl_data, size);
+			return;
 		}
-#else
-		cob_put_s64_compx (val, cbl_data, size);
 #endif
+		cob_put_s64_compx (val, cbl_data, size);
 		return;
-
 	case COB_TYPE_NUMERIC_PACKED:
 		cob_put_s64_comp3 (val, cbl_data, size);
 		return;
-
 	case COB_TYPE_NUMERIC_FLOAT:
-		flt = (float)val;  // possible data loss is explicit requested
+		flt = (float)val;  /* possible data loss is explicit requested */
 		cob_put_comp1 (flt, cbl_data);
 		return;
-
 	case COB_TYPE_NUMERIC_DOUBLE:
-		dbl = (double)val; // possible data loss is explicit requested
+		dbl = (double)val; /* possible data loss is explicit requested */
 		cob_put_comp2 (dbl, cbl_data);
 		return;
 	default:	/* COB_TYPE_NUMERIC_EDITED, ... */
@@ -1813,41 +2130,36 @@ cob_put_u64_param (int n, cob_u64_t val)
 		return;
 	}
 
-	cbl_data = f->data;
-	size = f->size;
 	if (COB_FIELD_CONSTANT (f)) {
-		cob_runtime_warning (_("%s: attempt to over-write constant param %d with " CB_FMT_LLD),
-							"cob_put_u64_param", n, val);
+		cob_runtime_warning_external ("cob_put_u64_param", 1,
+			_("attempt to over-write constant parameter %d with " CB_FMT_LLD),
+			n, val);
 		return;
 	}
+	cbl_data = f->data;
+	size = f->size;
 	switch (f->attr->type) {
 	case COB_TYPE_NUMERIC_DISPLAY:
 		cob_put_u64_pic9 (val, cbl_data, size);
 		return;
-
 	case COB_TYPE_NUMERIC_BINARY:
 #ifndef WORDS_BIGENDIAN
-		if (COB_FIELD_BINARY_SWAP (f)) {
-			cob_put_u64_compx (val, cbl_data, size);
-		} else {
+		if (!COB_FIELD_BINARY_SWAP (f)) {
 			cob_put_u64_comp5 (val, cbl_data, size);
+			return;
 		}
-#else
-		cob_put_u64_compx (val, cbl_data, size);
 #endif
+		cob_put_u64_compx (val, cbl_data, size);
 		return;
-
 	case COB_TYPE_NUMERIC_PACKED:
 		cob_put_u64_comp3 (val, cbl_data, size);
 		return;
-
 	case COB_TYPE_NUMERIC_FLOAT:
-		flt = (float)val;  // possible data loss is explicit requested
+		flt = (float)val;  /* possible data loss is explicit requested */
 		cob_put_comp1 (flt, cbl_data);
 		return;
-
 	case COB_TYPE_NUMERIC_DOUBLE:
-		dbl = (double)val;  // possible data loss is explicit requested
+		dbl = (double)val;  /* possible data loss is explicit requested */
 		cob_put_comp2 (dbl, cbl_data);
 		return;
 	default:	/* COB_TYPE_NUMERIC_EDITED, ... */
@@ -1870,8 +2182,9 @@ cob_put_picx_param (int n, void *char_field)
 	}
 
 	if (COB_FIELD_CONSTANT (f)) {
-		cob_runtime_warning (_("%s: attempt to over-write constant param %d with '%s'"),
-						"cob_put_picx_param", n, (char*)char_field);
+		cob_runtime_warning_external ("cob_put_picx_param", 1,
+			_("attempt to over-write constant parameter %d with '%s'"),
+			n, (char*)char_field);
 		return;
 	}
 
@@ -1879,15 +2192,14 @@ cob_put_picx_param (int n, void *char_field)
 }
 
 void *
-cob_get_grp_param (int n, void *char_field, int len)
+cob_get_grp_param (int n, void *char_field, size_t len)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_get_grp_param");
 
 	if (f == NULL) {
 		return NULL;
 	}
-
-	if (len <= 0) {
+	if (len == 0) {
 		len = f->size;
 	}
 
@@ -1902,7 +2214,7 @@ cob_get_grp_param (int n, void *char_field, int len)
 }
 
 void
-cob_put_grp_param (int n, void *char_field, int len)
+cob_put_grp_param (int n, void *char_field, size_t len)
 {
 	cob_field	*f = cob_get_param_field (n, "cob_put_grp_param");
 
@@ -1911,11 +2223,12 @@ cob_put_grp_param (int n, void *char_field, int len)
 	}
 
 	if (COB_FIELD_CONSTANT (f)) {
-		cob_runtime_warning (_("%s: attempt to over-write constant param %d"), "cob_put_grp_param", n);
+		cob_runtime_warning_external ("cob_put_grp_param", 1,
+			"attempt to over-write constant parameter %d", n);
 		return;
 	}
 
-	if (len <= 0 || len > f->size) {
+	if (len == 0 || len > f->size) {
 		len = f->size;
 	}
 	memcpy (f->data, char_field, len);
